/*
 * CPT - a Constraint Programming Temporal planner
 *
 * File : propagations.c
 *
 * Copyright (C) 2005, 2006  Vincent Vidal <vidal@cril.univ-artois.fr>
 */


#include "cpt.h"
#include "structs.h"
#include "problem.h"
#include "propagations.h"
#include "resources.h"
#include "preprocess.h"
#include "scheduling.h"
#include "trace_planner.h"
#include "globs.h"


/*---------------------------------------------------------------------------*/
/* Local Macros                                                              */
/*---------------------------------------------------------------------------*/


#define make_precede_cond(a1, a2) \
  NEST(									\
       bool cqp = opt.complete_qualprec;				\
       if (a1 != a2 && !precedes(a1, a2) && ((!cqp && (a1)->used && (a2)->used) || (cqp && !(a1)->excluded && !(a2)->excluded))) make_precede(a1, a2); )

#define must_precede(a1, a2) (precedes(a1, a2) || last_start(a1) + delta_aa(a1, a2) <= first_start(a2))


/* #define make_order_ac(a, c) \ */
/*   NEST(									\ */
/*        if (a != (c)->consumer && ((!(a)->excluded && (c)->required) || ((a)->used && !(c)->excluded))) { \ */
/* 	 if (cannot_precede_ca(c, a)) order_before_ac(a, c);		\ */
/* 	 if (cannot_precede_ac(a, c)) order_before_ca(c, a); } ) */


/*---------------------------------------------------------------------------*/
/* Static Functions                                                          */
/*---------------------------------------------------------------------------*/


static void synchronize_causal(Causal *c);
static void protect_causal(Causal *c);
static void protect_against(Action *a);
static void require(Causal *c);
static void make_order_aa(Action *a1, Action *a2);
static void make_order_ac(Action *a, Causal *c);
static void protect(Action *a1, Action *a2);
static void clone_causal(Causal *clone, Causal *c);


/*****************************************************************************/


void propagate_action(Action *a)
{
  if (a->excluded) return;
  trace_proc(propagate_action, a);
  if (opt.bound > 0 && last_start(end_action) < MAXTIME && first_end(a) > opt.bound) { exclude_action(a); return; }
  protect_against(a);
  if (a->excluded) return;
  FOR(c, a->causals) { activate_causal(c); } EFOR;
  FORCOUPLE(f, a->add, c, f->causals) { if (can_produce(c, a)) activate_causal(c); } EFORCOUPLE;

  //if (a->synchro) propagate_resource(a); 
}

void propagate_causal(Causal *c)
{
  if (c->excluded) return;
  trace_proc(propagate_causal, c);
  synchronize_causal(c);
  if (!c->excluded) protect_causal(c);
  if (opt.unique_supports && c->required && actual_event == Instantiate && edeletes(c->consumer, c->fluent)) {
    Action *prod = get_producer(c);
    if (prod && prod->used) {
      FOR(c2, c->fluent->causals) {
	if (c != c2 && edeletes(c2->consumer, c->fluent)) {
	  //print_action(get_producer(c)); printf ("    "); print_causal(c); printf ("    "); print_causal(c2); printf ("\n"); 
	  rem_producer(c2, get_producer(c));
	}
      } EFOR;
    }
  }
  if (!c->excluded && c->consumer->synchro) propagate_resource(c->consumer); 
  if (c->consumer == end_action && c->fluent->resource && resource_local(c->consumer, c->fluent->resource)) protect_causal_resource(c);
}



#ifdef SCHED 

#define valdroite1(x) (x + (x - xa) * (yb - ya) / (xb - xa) + ya)
#define valdroite2(x) (x + (int) ((x - xa) * (yb - ya) / (double) (xb - xa) + 0.5) + ya)

void compute_ac_constraint(ActivityConstraint *al, Time fs, Time ls, Time *almin, Time *almax)
{
  *almin = al->min;
  *almax = al->max;
  if (al->time) {
    Time mi = MAXTIME, ma = 0, xa, ya, xb, yb;
    Num i;
    minimize(ls, last_start(end_action));
    for (i = 0; i < al->time_nb - 1; i++) {
      xa = al->time[i];
      ya = al->dur[i];
      xb = al->time[i + 1];
      yb = al->dur[i + 1];
      if (fs <= al->time[i + 1]) {
	if (fs >= al->time[i]) { minimize(mi, valdroite1(fs)); maximize(ma, valdroite2(fs)); }
	if (ls <= al->time[i + 1]) { minimize(mi, valdroite1(ls)); maximize(ma, valdroite2(ls)); break; }
	minimize(mi, valdroite1(al->time[i + 1]));
	maximize(ma, valdroite2(al->time[i + 1]));
      }
    }
    *almin += mi - fs;
    *almax += ma - ls;
/*     printf("ICI : min = %d max = %d -- fs = %d ls = %d\n", *almin, *almax, fs, ls); */
  }
}

#define valdroite5(x) ((x - xa) * (yb - ya) / (xb - xa) + ya)
#define valdroite6(x) ((int) ((x - xa) * (yb - ya) / (double) (xb - xa) + 0.5) + ya)

void compute_ac_constraint2(ActivityConstraint *al, Time fs, Time ls, Time *almin, Time *almax)
{
  if (al->time) {
    Time xa, ya, xb, yb;
    Num i;
    *almin = last_start(end_action);
    *almax = 0;
    for (i = 0; i < al->time_nb - 1; i++) {
      xa = al->time[i];
      ya = xa + al->dur[i];
      xb = al->time[i + 1];
      yb = xb + al->dur[i + 1];
      if (xa <= fs && fs <= xb) { minimize(*almin, valdroite5(fs)); maximize(*almax, valdroite6(fs)); }
      if (fs <= xb && xb < ls) { minimize(*almin, yb); maximize(*almax, yb); }
      if (ls <= xb) { minimize(*almin, valdroite5(ls)); maximize(*almax, valdroite6(ls)); break; }
    } 
    *almin += al->min;
    *almax += al->max;
    //printf("ICI : min = %d max = %d -- fs = %d ls = %d\n", *almin, *almax, fs, ls); fflush(stdout);
  } else {
    *almin = fs + al->min;
    *almax = ls + al->max;
  }
}

#define valdroite3(x) ((x - ya) * (xb - xa) / (yb - ya) + xa)
#define valdroite4(x) ((int) ((x - ya) * (xb - xa) / (double) (yb - ya) + 0.5) + xa)

void compute_ac_constraint3(ActivityConstraint *al, Time fs, Time ls, Time *almin, Time *almax)
{
  if (al->time) {
    Time xa, ya, xb, yb;
    Num i;
    minimize(ls, last_start(end_action));
    *almin = last_start(end_action);
    *almax = 0;
    fs -= al->max;
    ls -= al->min;
    for (i = 0; i < al->time_nb - 1; i++) {
      xa = al->time[i];
      ya = xa + al->dur[i];
      xb = al->time[i + 1];
      yb = xb + al->dur[i + 1];
      if (fs <= ya) { *almin = xa; break; }
      if (fs <= yb) { *almin = valdroite3(fs); break; }
    }

    for (i = al->time_nb - 2; i>= 0; i--) {
      xa = al->time[i];
      ya = xa + al->dur[i];
      xb = al->time[i + 1];
      yb = xb + al->dur[i + 1];
      //printf("%d %d %d %d\n", xa, ya, xb, yb); 
      if (ls >= yb) { *almax = xb; break; }
      if (ls >= ya) { *almax = valdroite4(ls);  break; }
    }
/*     *almin -= al->max; */
/*     *almax -= al->min; */
    //printf("ICI : min = %d max = %d -- fs = %d ls = %d\n", *almin, *almax, fs, ls); fflush(stdout);
  } else {
    *almin = fs - al->max;
    *almax = ls - al->min;
  }
}

#endif


static void synchronize_causal(Causal *c)
{
  Action *prod = get_producer(c);
  
  trace_proc(synchronize_causal, c);
  if (prod) {
    trace_proc(producer_used, c); 
    if (!prod->used && prod->origin->nb_instances > 0) clone_action(prod, c);
#ifdef SCHED
    FOR(al, prod->ac_constraints) {
      if (al->fluent == c->fluent) {
	Time almin, almax;

	compute_ac_constraint3(al, first_start(c->consumer), last_start(c->consumer), &almin, &almax);
	update_sup_a(prod, almax - duration(prod));
	update_inf_a(prod, almin - duration(prod));
	compute_ac_constraint2(al, first_end(prod), last_end(prod), &almin, &almax);
	update_inf_a(c->consumer, almin);
	update_sup_a(c->consumer, almax);
	break;
      }
    } EFOR;
#endif
    update_inf_a(prod, first_start(c));
    update_inf_c(c, first_start(prod));
    update_inf_a(c->consumer, first_start(prod) + delta_aa(prod, c->consumer));
    update_sup_a(prod, min(last_start(c), last_start(c->consumer) - delta_aa(prod, c->consumer)));
    update_sup_c(c, last_start(prod));
    use_action(prod);
    if (c->consumer->used) make_precede_cond(prod, c->consumer);
  } else {
    Time min_cons_init = MAXTIME, min_prod_init = MAXTIME, max_prod_init = 0;
#ifdef SCHED
    Time max_cons_init = 0;
#endif

    FORPROD(a, c) {
      if (last_start(a) < first_start(c) || first_start(a) > last_start(c) || cannot_precede_aa(a, c->consumer) || first_start(a) > last_start(end_action))
	rem_producer(c, a); 
      else {
#ifdef SCHED
	Time almin, almax;
	
	FOR(al, a->ac_constraints) {
	  if (al->fluent == c->fluent) {
	    compute_ac_constraint(al, first_end(a), last_end(a), &almin, &almax);
	    minimize(min_cons_init, max(first_start(a), first_start(c)) + max(duration(a) + almin, delta_aa(a, c->consumer)));
	    minimize(min_prod_init, max(first_start(a), first_start(c->consumer) - duration(a) - almax));
	    maximize(max_prod_init, min(last_start(a), last_start(c->consumer) - max(duration(a) + almin, delta_aa(a, c->consumer))));
	    if (MAXTIME - last_start(a) >= duration(a) + almax) maximize(max_cons_init, last_end(a) + almax);
	    else max_cons_init = MAXTIME;
	    goto ac_done;
	  }
	} EFOR;
	max_cons_init = MAXTIME;
#endif
	minimize(min_cons_init, max(first_start(a), first_start(c)) + delta_aa(a, c->consumer));
	minimize(min_prod_init, first_start(a));
	maximize(max_prod_init, min(last_start(a), last_start(c->consumer) - delta_aa(a, c->consumer)));
#ifdef SCHED
      ac_done:;
#endif
      }
    } EFOR;
    update_inf_a(c->consumer, min_cons_init);
    update_inf_c(c, min_prod_init);
    update_sup_c(c, max_prod_init);
#ifdef SCHED
    update_sup_a(c->consumer, max_cons_init);
#endif
  }
}

#ifdef SCHED

static void protect_causal(Causal *c)
{
  //return;
  if (c->required || opt.propagate_inactive_causals) {
    trace_proc(protect_causal, c);
    FOR(a, c->fluent->edeleters) { if (!a->excluded) make_order_ac(a, c); if (c->excluded) return; } EFOR;
  }
  if (!c->excluded && opt.local_mutex_sets) local_mutex_sets(c);
}

static void protect_against(Action *a)
{
  //return;
  trace_proc(protect_against, a);

/*   if (a->used) { */
/*     FORCOUPLE(f, a->add, a2, f->edeleters) { if (a2->used) { make_order_aa(a, a2); } } EFORCOUPLE; */
/*     FORCOUPLE(f, a->edel, a2, f->producers) { if (a2->used) make_order_aa(a, a2); } EFORCOUPLE; */
/*   } */
  FORCOUPLE(f, a->edel, c, f->active_causals) { 
    make_order_ac(a, c); if (a->excluded) return; 
  } EFORCOUPLE;
  if (a->used) {
/*     FOR(a2, active_actions) { make_order_aa(a, a2); } EFOR; */
    FOR(a2, active_actions) {
      if (a != a2) {
	if (precedes(a, a2)) order_before_aa(a, a2);
	else if (precedes(a2, a)) order_before_aa(a2, a);
      }
    } EFOR;
  }
}

#else

static void protect_causal(Causal *c)
{
  if (c->required || opt.propagate_inactive_causals) {
    trace_proc(protect_causal, c);
    FOR(a, c->fluent->edeleters) { if (!a->excluded) make_order_ac(a, c); if (c->excluded) return; } EFOR;
  }
  if (!c->excluded && opt.local_mutex_sets) local_mutex_sets(c);
}

static void protect_against(Action *a)
{
  trace_proc(protect_against, a);

  if (a->used) {
    FORCOUPLE(f, a->add, a2, f->edeleters) { if (a2->used) make_order_aa(a, a2); } EFORCOUPLE;
    FORCOUPLE(f, a->edel, a2, f->producers) { if (a2->used) make_order_aa(a, a2); } EFORCOUPLE;
  }
  if (opt.landmarks) {
    FOR(a2, a->after) { order_before_aa(a, a2); } EFOR;
    FOR(a2, a->before) { order_before_aa(a2, a); } EFOR;
  }
  if (opt.propagate_inactive_causals) 
    FORCOUPLE(f, a->edel, c, f->causals) { make_order_ac(a, c); if (a->excluded) return; } EFORCOUPLE;
  else 
    FORCOUPLE(f, a->edel, c, f->active_causals) { make_order_ac(a, c); if (a->excluded) return; } EFORCOUPLE;
  if (opt.sequential) {
    if (a->used) {
      FOR(a2, actions) {
	if (a != a2) {
	  if (cannot_precede_aa(a, a2)) order_before_aa(a2, a);
	  if (cannot_precede_aa(a2, a)) order_before_aa(a, a2);
	}
      } EFOR;
    } else {
      FOR(a2, active_actions) {
	if (cannot_precede_aa(a, a2)) order_before_aa(a2, a);
	if (cannot_precede_aa(a2, a)) order_before_aa(a, a2);
      } EFOR;
    }
  }
  if (a->used) {
    FOR(a2, active_actions) {
      if (a != a2) {
	if (precedes(a, a2)) order_before_aa(a, a2);
	else if (precedes(a2, a)) order_before_aa(a2, a);
      }
    } EFOR;
  }

  if (!opt.optimal) goto test;
  return;
  FOR(c, a->causals) {
    FORPROD(a2, c) {
      FOR(f, a2->add) {
	if (f != c->fluent && edeletes(a, f) && !deletes(a, f)) {
	  FOR(del, f->deleters) {
	    if (!del->excluded && !edeletes(del, c->fluent) && !cannot_precede_aa(a2, del) && !cannot_precede_aa(del, a)) goto suite2;
	  } EFOR;
/* 	  print_action(a2); */
/* 	  printf(" cannot produce "); */
/* 	  print_causal(c); */
/* 	  printf("\n"); */
	  rem_producer(c, a2);
	  goto suite3;
	suite2:;
	}
      } EFOR;
    suite3:;
    } EFOR;
  } EFOR;

  //return;

 test:
  if (opt.landmarks && a->used) {
    FOR(f, a->add) {
      FOR(c, active_causals) {
	if (precedes(a, c->consumer) && edeletes(c->consumer, f) && !deletes(c->consumer, f)) {
	  Action *d = NULL;
	  FOR(del, f->deleters) {
	    if (!del->excluded && !precedes(del, a) && !precedes(c->consumer, del)) {
	      if (d == NULL) d = del;
	      else goto suite;
	    }
	  } EFOR;
	  if (!d) contradiction();
/* 	  printf("%s devra etre dtruit ", fluent_name(f)); */
/* 	  printf("par %s\n", action_name(d)); */
	  if (!d->used) {
	    clone_action(d, NULL);
	    use_action(d);
	  }
	  FOR(cc, d->causals) {
	    if (cc->fluent == f && can_produce(cc, a)) {
	      set_producer(cc, a);
	      break;
	    }
	  } EFOR;
	suite:;
	}
      } EFOR;
    } EFOR;
  }
}
#endif

void use_action(Action *a)
{
  if (!a->used) {
    store(a->used, true);
    trace_proc(action_used, a);
    FOR(c, a->causals) { rem_producer(c, a); require(c); } EFOR;
    store_tab(active_actions, a);
    if (opt.complete_qualprec) {
      FOR(a1, actions) {
	if (!a1->excluded && precedes(a1, a) && a1 != a)
	  FOR(a2, actions) {
	    if (!a2->excluded && precedes(a, a2) && !precedes(a1, a2)) 
	      order_before_aa(a1, a2);
	  } EFOR;
      } EFOR;
      
      FORCOUPLE(f, a->edel, c, f->causals) {
	if (!c->consumer->excluded && precedes(a, c->consumer))
	  FORPROD(prod, c) {
	    if (precedes(prod, a)) rem_producer(c, prod);
	  } EFOR;
      } EFORCOUPLE;
    }
  }
  protect_against(a); 
}


#define increment_weight(c) NEST( c->origin->weight++; maximize(c->consumer->origin->weight, c->origin->weight); )

static void exclude_action2(Action *a, Causal *c)
{
  if (a->excluded) return;
  if (a == end_action && !a->used) return;
  trace_proc(action_excluded, a);
  if (a->used) { 
    if (c) {
      increment_weight(c);
    } else {
      FORCOUPLE(f, a->add, c, f->active_causals) { if (get_producer(c) == a) increment_weight(c); } EFORCOUPLE;
    }
    //FOR(b, active_actions) { if (b != a && amutex(a, b) && (precedes(a, b) || precedes(b,a))) b->weight++; } EFOR;
    FOR(b, active_actions) { if (b != a && amutex(a, b)) b->weight++; } EFOR;
    a->weight++;
    contradiction(); 
  }
  store(a->excluded, true);
  FORCOUPLE(f, a->add, c, f->causals) { rem_producer(c, a); } EFORCOUPLE;
  FOR(c, a->causals) { store(c->excluded, true); } EFOR;
}

void exclude_causal(Causal *c)
{
  exclude_action2(c->consumer, c);
}


void exclude_action(Action *a)
{
  exclude_action2(a, NULL);
}


static void require(Causal *c)
{
  store(c->required, true);
  store_tab(active_causals, c);
  store_tab(c->fluent->active_causals, c);
  activate_causal(c);
}

static void make_order_aa(Action *a1, Action *a2)
{
  if (a1 != a2 && ((!(a1)->excluded && a2->used) || (a1->used && !(a2)->excluded))) {
    if (cannot_precede_aa(a2, a1)) order_before_aa(a1, a2);
    else if (cannot_precede_aa(a1, a2)) order_before_aa(a2, a1);
  }
}

static void make_order_ac(Action *a, Causal *c)
{
  if (a != (c)->consumer && ((!(a)->excluded && c->required) || (a->used && !(c)->excluded))) {
    if (cannot_precede_ca(c, a)) order_before_ac(a, c);
    if (cannot_precede_ac(a, c)) order_before_ca(c, a);
  }
}

void order_before_aa(Action *a1, Action *a2)
{
  Time d = delta_aa(a1, a2);

  trace_proc(order_before_aa, a1, a2);
  if (a1->used) update_inf_a(a2, first_start(a1) + d);
  if (a2->used) update_sup_a(a1, last_start(a2) - d);
  if (!precedes(a1, a2) && (!duration(a1) || !duration(a2))) FOR(c, a1->causals) { rem_producer(c, a2); } EFOR; // rajout pour actions de synchro
  make_precede_cond(a1, a2);
}

void order_before_ca(Causal *c, Action *a)
{
  trace_proc(order_before_ca, c, a);
  order_before_aa(c->consumer, a);
}

void order_before_ac(Action *a, Causal *c)
{
  Action *prod = get_producer(c);

  trace_proc(order_before_ac, a, c);
  if (prod) {
    if (a != prod && prod->used) order_before_aa(a, prod);
  } else {
    if (can_produce(c, a)) {
      if (c->required) update_sup_a(a, last_start(c));
      if (a->used) update_inf_c(c, first_start(a));
    } else {
      Time d = delta_ac(a, c);
      if (c->required) update_sup_a(a, last_start(c) - d);
      if (a->used) update_inf_c(c, first_start(a) + d);
    }
    make_precede_cond(a, c->consumer);
  }
}

void make_precede(Action *a1, Action *a2)
{
  Action **act;
  Num act_nb;
  
  if  (opt.complete_qualprec) { act = actions; act_nb = actions_nb; }
  else { act = active_actions; act_nb = active_actions_nb; }
  set_precedes(a1, a2);
  protect(a1, a2);
  FOR(x, act) {
    if (!x->excluded && x != a1 && x != a2) {
      if (a2->used && precedes(a2, x) && !precedes(a1, x)) {
	order_before_aa(a1, x);
	if (a1->excluded) return;
	if (x->excluded) continue;
/* 	set_precedes(a1, x); */
/* 	protect(a1, x); */
/* 	activate_action(x); */
      }
      if (a1->used && precedes(x, a1) && !precedes(x, a2)) {
	order_before_aa(x, a2);
	if (a2->excluded) return;
	if (x->excluded) continue;
/* 	set_precedes(x, a2); */
/* 	protect(x, a2); */
/* 	activate_action(x); */
      }
    }
  } EFOR;
  activate_action(a1);
  activate_action(a2);
  protect(a1, a2);
}

static void protect(Action *a1, Action *a2)
{
  Causal **caus;
  Num caus_nb;

  if (precedes(a2, a1)) {
    if (a1->used) exclude_action(a2);
    else if (a2->used) exclude_action(a1);
  } else {
    if (a2->used)
      FOR(f, a1->add) {
	if (edeletes(a2, f)) {
	  if  (opt.complete_qualprec) { caus = f->causals; caus_nb = f->causals_nb; }
	  else { caus = f->active_causals; caus_nb = f->active_causals_nb; }
	  FOR(c, caus) {
	    if (precedes(a2, c->consumer)) {
	      rem_producer(c, a1);
	    }
	  } EFOR;
	}
      } EFOR;
    
    if (a1->used)
      FOR(c, a2->causals) {
	if (edeletes(a1, c->fluent))
	  FORPROD(prod, c) {
	    if (precedes(prod, a1))
	      rem_producer(c, prod);
	  } EFOR;
      } EFOR;
  }
}

static void clone_causal(Causal *clone, Causal *c)
{
  memcpy(clone, c, CSIZE_CAUSAL); 
  clone_variable(&clone->start, &c->start); 
  clone_variable(&clone->support, &c->support);
  store(causals_nb, causals_nb + 1);
  store_tab(clone->fluent->causals, clone);
}

void clone_action(Action *a, Causal *causal)
{
  Action *clone = actions[actions_nb];

  if (!clone) error(max_plan_length, "Increase the maximum plan length (actually %ld)", opt.max_plan_length);
  store(a->origin->nb_instances, a->origin->nb_instances - 1);
  memcpy(clone, a, CSIZE_ACTION); 
  clone_variable(&clone->start, &a->start);
  memcpy(clone->properties, a->properties, properties_nb * sizeof(Property));
  store(actions_nb, actions_nb + 1);
  if (a->synchro) clone_variable(&clone->reslevel, &a->reslevel);

  FOR2(c1, clone->causals, c2, a->causals) { clone_causal(c1, c2); } EFOR;    

  FOR(f, a->add) {
    FOR(c, f->causals) {
      if (c != causal && can_produce(c, a)) {
	store(c->fluent->indac[clone->num], f->producers_nb);
	add_val_variable(&c->support, (Value) f->producers_nb);
      }
    } EFOR;
    store_tab(f->producers, clone);
  } EFOR;

  FOR(f, clone->del) { store_tab(f->deleters, clone); } EFOR;
  FOR(f, clone->edel) { store_tab(f->edeleters, clone); } EFOR;
  FOR(f, clone->prec) { store_tab(f->consumers, clone); } EFOR;
  if (opt.complete_qualprec) {
    FOR(a2, actions) {
      if (precedes(a2, a)) set_precedes(a2, clone);
      if (precedes(a, a2)) set_precedes(clone, a2);
    } EFOR;
  }
  activate_action(clone);
}


bool preserves(Action *a, Causal *cinit, Action *b, Causal *c, Fluent *f)
{
  if (deletes(c->consumer, f) && !produces(c->consumer, f)) return false;
  FOR(del, f->deleters) {
    if (!del->excluded && del != c->consumer && del != b/*  && !edeletes(del, c->fluent)*/ && !must_precede(del, b) && !must_precede(c->consumer, del)) return false;
  } EFOR;

/*   printf("\n"); */
/*   print_action(a); */
/*   printf(" --> "); */
/*   print_causal(c); */
/*   printf("implies %s", fluent_name(f)); */
/*   printf("\n"); */
  
  return true;

  if (get_producer(cinit) == a) {
    FOR(f2, fluents) {
      if (f!= f2 && fmutex(f, f2) && !edeletes(c->consumer, f2)) { // && !produces(c->consumer, f2)) {
	//print_action(c->consumer); printf(" edeletes %s\n", fluent_name(f2));
	set_temp_edeletes(c->consumer, f2);
	store_tab(f2->edeleters, c->consumer);
	if (c->consumer->num != 1) store_tab(c->consumer->edel, f2);
	activate_action(c->consumer);
      }
    } EFOR;
  }
  
  return true;
}

bool protect_interval(Action *a, Causal *c, Action *b, Fluent *f)
{
  FOR(a2, f->edeleters) {
    if (a2 != a && b != a2 && !a2->excluded && a2 != c->consumer && !produces(a2, f)) {
      if (0&&get_producer(c) == a) {
	if (cannot_precede_aa(b, a2)) order_before_aa(a2, a);
	else if (cannot_precede_aa(a2, a)) order_before_aa(b, a2);
      } else if ((precedes(a, a2) || precedes(c->consumer, a2)) && precedes(a2, b)) {
/*       } else if (cannot_precede_aa(b, a2) && (cannot_precede_aa(a2, a) || cannot_precede_ac(a2, c))) { */
/* 	printf("\n%s%s cannot support ", (a->used? "*" : ""), action_name(a)); */
/* 	print_causal(c); */
/* 	printf("\ncause : "); */
/* 	print_action(a2); */
/* 	printf("\n"); */
	if (a2->used) return true;
      }
    }
  } EFOR;
  return false;
}

bool pres(Action *a, Causal *cinit, Action *b, Causal *c, Fluent *f, int level)
{
  if (preserves(a, cinit, b, c, f)) {
    if (edeletes(c->consumer, f)/*  && !deletes(c->consumer, f) */) { 
      //printf("%d ", level); 
      return true;
    }
    if  (protect_interval(a, cinit, c->consumer, f)) {
      //printf("%d ", level); 
      return true;
    }
    FOR(f2, c->consumer->add) {
      bool res = false;
      FOR(c2, f2->active_causals) {
	if (c->consumer == get_producer(c2)) {
	  if (pres(a, cinit, c->consumer, c2, f, level+1)) return true; //res = true;
/* 	  if (pres(a, cinit, c2, f, level+1)) res = true; */
/* 	} else if (get_producer(cinit) == a && can_produce(c2, c->consumer)) { */
/* 	  if (preserves(a, c2, f)) { */
/* 	      if (edeletes(c2->consumer, f) && !deletes(c2->consumer, f)) {  */
/* 		rem_producer(c2, c->consumer); */
/* 	      } */
/* 	    } */
	}
      } EFOR;
      if (res) return true;
    } EFOR;
/*   } else { */
/*     if  (protect_interval(a, cinit, b, f)) { */
/*       //printf("%d ", level);  */
/*       return true; */
/*     } */
  }
  return false;
}


void more_propagations(void)
{
  if (opt.global_mutex_sets) global_mutex_sets();
  if (opt.task_intervals) task_intervals();

/*   FOR(a, actions) { */
/*     if (a->origin->num==140) { */
/*       print_action(a); */
/*       printf("   %ld %ld\n", min_level(a), max_level(a)); */
/*     } */
/*   } EFOR; */


  return;
#ifdef SCHED
  FOR(a, actions) {
    if (a->excluded) goto cont;
/*     FOR(c, a->causals) { */
/*       FOR(a2, c->fluent->edeleters) { make_order_ac(a2, c); if (a->excluded) goto cont; } EFOR; */
/*     } EFOR; */
    FORCOUPLE(f, a->edel, c, f->causals) { 
      make_order_ac(a, c); if (a->excluded) goto cont; 
    } EFORCOUPLE;
/*     if (a->used) { */
/*       FOR(a2, active_actions) { */
/* 	if (a != a2) { */
/* 	  if (precedes(a, a2)) order_before_aa(a, a2); */
/* 	  else if (precedes(a2, a)) order_before_aa(a2, a); */
/* 	} */
/*       } EFOR; */
/*     } */

    if (a->used) FOR(a2, active_actions) { make_order_aa(a, a2); if (a->excluded) goto cont; } EFOR;
/*     FOR(a2, actions) { make_order_aa(a, a2); if (a->excluded) goto cont; } EFOR; */
  cont:;
  } EFOR;
#endif

/*   compute_bad_supporters3(); */
  return;
  goto test;
  FOR(c, causals) {
    Action *a = get_producer(c);
    if (a) {
      FOR(a2, active_actions) {
	if (a2 != a && a2 != c->consumer && precedes(a, a2) && precedes(a2, c->consumer)) {
	  FOR(f, a2->edel) {
	    if (!deletes(a2, f) && !deletes(a, f) && !edeletes(a, f)) {
	      FOR(del, f->deleters) {
		if (!del->excluded && !edeletes(del, c->fluent) && !must_precede(del, a) && !must_precede(a2, del)) goto suite;
	      } EFOR;
/* 	      print_action(a); */
/* 	      printf(" edeletes %s\n", fluent_name(f)); */
	      set_temp_edeletes(a, f);
	      store_tab(f->edeleters, a);
	      if (c->consumer->num != 1) store_tab(a->edel, f);
	      activate_action(a);
	    suite:;
	    }
	  } EFOR;
	}
      } EFOR;
    }
  } EFOR;

  return;
 test:

  FOR(c, active_causals) {
    FORPROD(a, c) {
      FOR(f, a->add) {
	if (pres(a, c, a, c, f, 1)) {
	  rem_producer(c, a);
	  break;
	}
      } EFOR;
    } EFOR;
  } EFOR;

}

void restrict_init_state(Fluent **init, Num init_nb, bool prop)
{
  if (init_nb != -1) {
    start_action->add = init;
    start_action->add_nb = init_nb;
    FOR(f, fluents) { f->is_init = false; } EFOR;
    FOR(f, init) { f->is_init = true; } EFOR;
    FOR(f, fluents) {
      if (!f->is_init) {
	FOR(c, f->causals) {
	  rem_producer(c, start_action);
	} EFOR;
      }
    } EFOR;
  }
  if (prop) propagate_init_state();
}

static Causal **end_action_causals;
static int end_action_causals_nb;

void restrict_goal_state(Fluent **goal, Num goal_nb, bool prop)
{
  if (goal_nb != -1) {
    if (end_action_causals == NULL) {
      my_malloc(end_action_causals, end_action->causals_nb);
      memcpy(end_action_causals, end_action->causals, sizeof(Causal *) * end_action->causals_nb);
      end_action_causals_nb = end_action->causals_nb;
    }
    end_action->prec = goal;
    end_action->prec_nb = goal_nb;
    FOR(f, fluents) { f->is_init = false; } EFOR;
    FOR(f, goal) { f->is_init = true; } EFOR;
    end_action->causals_nb = 0;
    FOR(c, end_action_causals) {
      if (c->fluent->is_init) {
	end_action->causals[end_action->causals_nb++] = c;
      }
      else store(c->excluded, true);
    } EFOR;
  }
  update_sup_a(end_action, opt.max_makespan);
  if (prop) propagate_goal_state();
}

void propagate_init_state(void)
{
  use_action(start_action);
  update_inf_a(start_action, 0);
  update_sup_a(start_action, 0);
  make_precede(start_action, end_action);
  
  FORMIN(a, actions, 2) {
    make_precede(start_action, a);
    make_precede(a, end_action);
  } EFOR;
  
/*   FORMINMAX(a, actions, actions_nb, actions_nb + opt.max_plan_length) { */
/*     set_precedes(start_action, a); */
/*     set_precedes(a, end_action); */
/*   } EFOR;     */
  
  propagate();

  if (!end_action->used) set_min(&end_action->start, 0);
}

void propagate_goal_state(void)
{
  use_action(end_action);
  propagate();
}
